/*
 * Segment 2 (KEx),  SWE 2.DE Statischer Import MIF
 * Copyright (C) 2007 BitCtrl Systems GmbH 
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.kex.mif.konfigdaten;

import java.util.HashSet;

import org.apache.xerces.util.XMLChar;
import org.geotools.feature.FeatureType;

import com.vividsolutions.jts.geom.Geometry;

import de.bsvrz.dav.daf.main.ClientDavConnection;
import de.bsvrz.dav.daf.main.config.SystemObjectType;
import de.bsvrz.kex.mif.kb.KonfigurationsObjekt;
import de.bsvrz.kex.mif.mif.MifObjekt;
import de.bsvrz.kex.mif.mif.MifTyp;
import de.bsvrz.kex.mif.vew.MIFProperties;

/**
 * Factory-Klasse zur Erzeugung von Konfigurationsobjekten.
 *
 * @author BitCtrl Systems GmbH, Gieseler
 */
public final class MifGeoObjektFactory {
	/**
	 * Die (einzige) Instanz dieser Klasse.
	 */
	private static MifGeoObjektFactory instance;

	/**
	 * Gibt die Instanz dieser KLasse zur&uuml;ck.
	 *
	 * @param dav
	 *            {@link ClientDavConnection}
	 * @return {@link MifGeoObjektFactory}
	 */
	public static MifGeoObjektFactory getInstance(final ClientDavConnection dav) {
		if (MifGeoObjektFactory.instance == null) {
			MifGeoObjektFactory.instance = new MifGeoObjektFactory(dav);
		}

		return MifGeoObjektFactory.instance;
	}

	/**
	 * Datenverteiler-Verbindung.
	 */
	private final ClientDavConnection dav;

	/**
	 * Menge der PIDs.
	 */
	private final HashSet<String> pids = new HashSet<String>();

	/**
	 * Privater Konstruktor.
	 *
	 * @param dav
	 *            Datenverteiler-Verbindung.
	 */
	private MifGeoObjektFactory(final ClientDavConnection dav) {
		this.dav = dav;
	}

	/**
	 * Aktualisiert die PID durch Anf&uuml;gen eines Postfix falls die PID schon
	 * exitiert.
	 * 
	 * @param pid
	 *            PID
	 * @return PID oder korrigierte PID
	 */
	private String aktualisierePID(final String pid) {
		String newpid = pid;
		if (pids.contains(pid)) {
			int postfix = 1;
			do {
				// newpid = pid+"."+lfd_nr; //$NON-NLS-1$
				newpid = pid + postfix;
				postfix++;
			} while (pids.contains(newpid));
		}

		return newpid;
	}

	/**
	 * Bestimmt den Zieltyp im Datenkatalog f&uuml;r den Typ eines MifObjektes.
	 * Zielobjekttyp ist dabei ein vom Typ GeoReferenzObjekt abgeleiteter
	 * Subtyp, der die Zielobjekte repr&auml;sentiert. Der Zieltyp wird wie
	 * folgt gebildet: Wenn der gew&uuml;nschte Zieltyp ein Subtyp des zum
	 * MifTyp definierten Basistyps ist, wird der gew&uuml;nschte Zieltyp
	 * zur&uuml;ckgegeben, ansonsten der Basistyp. Wenn der gew&uuml;nschte
	 * Zieltyp nicht gesetzt ist, wird der Basistyp zur&uuml;ckgegeben.
	 *
	 * @param miftyp
	 *            {@link MifTyp}
	 * @param zieltyp
	 *            gew&uuml;nschter Zieltyp
	 * @return Typname GeoReferenzObjekt
	 */
	private String bestimmeDavTyp(final MifTyp miftyp, final String zieltyp) {
		if (zieltyp == null) {
			return miftyp.getDavTyp();
		}

		final SystemObjectType davtypziel = dav.getDataModel().getType(zieltyp);
		final SystemObjectType davtypmif = dav.getDataModel().getType(miftyp.getDavTyp());

		// alles kann auf "typ.komplexXY" abgebildet werden
		final SystemObjectType davtypzielkomplex = dav.getDataModel().getType("typ.komplexXY");
		if (davtypziel.getSuperTypes().contains(davtypzielkomplex)) {
			return davtypziel.getPid();
		}

		if (davtypziel.getSuperTypes().contains(davtypmif)) {
			return davtypziel.getPid();
		}

		if (MIFProperties.getInstance().istAktiv(MIFProperties.PROP_KEY_IMPORTALL)) {
			return davtypmif.getPid();
		}

		return null;
	}

	/**
	 * Bildet den Namen des Konfigurationsobjektes.
	 *
	 * @param pidname
	 *            PID
	 * @return Name
	 */
	private String bildeNamen(final String pidname) {
		return pidname;
	}

	/**
	 * Erzeugt einen NCNAMEN aus einem String.
	 * 
	 * @param text
	 *            String
	 * @return konvertierter String
	 */
	private String bildeNCNamen(final String text) {
		if (text != null) {
			int length = 0;
			final String nametext = text.trim();
			length = nametext.length();
			String ncnametext = "";

			if (text.length() == 0) {
				return text;
			}

			if (!XMLChar.isNCNameStart(nametext.charAt(0))) {
				ncnametext = new String("_");
			} else {
				ncnametext = new String(String.valueOf(nametext.charAt(0)));
			}

			for (int i = 1; i < length; i++) {
				if (!XMLChar.isNCName(nametext.charAt(i))) {
					ncnametext = ncnametext.concat("_");
				} else {
					ncnametext = ncnametext.concat(String.valueOf(nametext.charAt(i)));
				}
			}

			return ncnametext;
		}

		return text;
	}

	/**
	 * F&uuml;r die erzeugten Objekte werden die PID nach folgenden Regeln
	 * gebildet F&uuml;r die PID ist eine der Spalten aus dem COLUMNSHeader der
	 * .mif-Datei anzugeben. Zus&auml;tzlich wird bei der PID abh&auml;ngig vom
	 * erzeugten Basis-Objekttyp das Pr&auml;fix "punktXY", "linieXY",
	 * "fl&auml;cheXY" oder "komplexXY" vorangestellt. Die Spalte, deren Name
	 * bei dem zu konvertierenden Typ mit einem der m&ouml;glichen Spaltennamen
	 * &uuml;bereinstimmt, wird verwendet. Stimmt keiner der vorgegebenen Namen
	 * mit einem Spaltennamen &uuml;berein oder wurden keine Vorgaben gemacht,
	 * so wird die erste Spalte als Grundlage f&uuml;r die PID bzw. den Namen
	 * verwendet. Die Eindeutigkeit der PID wird bei der Konvertierung
	 * &uuml;berpr&uuml;ft. Bei Verletzung wird die Eindeutigkeit durch
	 * Anf&uuml;gen einer fortlaufenden Nummer als Postfix der PID hergestellt.
	 * Beispiel: fl&auml;cheXY.Rhein
	 *
	 * Bildung des Namens F&uuml;r den Namen wird derselbe Eintrag aus dem
	 * COLUMNS-Header der .mif-Datei verwendet, der auch zur Bildung der PID
	 * verwendet wurde. Eine Ausnahme bilden MIF-Datenobjekte vom Typ Text. Bei
	 * diesen Objekten wird der Wert des Objektes (der Text) als Name benutzt.
	 *
	 * Beispiel: Rhein
	 * 
	 * @param pidname
	 *            PID
	 * @param prefix
	 *            Prefix
	 * @return PID
	 */
	private String bildePID(final String pidname, final String prefix) {

		// alle Leerzeichen und ',' durch '_' ersetzen
		String korrpid = pidname.replaceAll("[ ,]", "_"); //$NON-NLS-1$ //$NON-NLS-2$

		// valid NCNAME sicherstellen
		if (!XMLChar.isValidNCName(korrpid)) {
			korrpid = bildeNCNamen(korrpid);
		}

		final String pid = prefix + "." + korrpid; //$NON-NLS-1$
		return aktualisierePID(pid);
	}

	/**
	 * Erzeugt ein Konfigurationsobjekt.
	 *
	 * @param mo
	 *            {@link MifObjekt} das zugeh&ouml;hrige Objekt der
	 *            MIF-Datenbasis.
	 * @param pidspaltenname
	 *            der Name der zur Bildeung der PID zu benutzenden Spalte der
	 *            MIF-Datenbasis
	 * @param davzieltyp
	 *            Zielttyp im DAV
	 * @return {@link KonfigurationsObjekt}
	 * @throws ImportException
	 *             bei Ausnahmen
	 */
	public KonfigurationsObjekt erzeugeMifGeoObjekt(final MifObjekt mo, final String pidspaltenname,
			final String davzieltyp) throws ImportException {
		final String davtyp = bestimmeDavTyp(mo.getMifTyp(), davzieltyp);

		if (davtyp == null) {
			// _debug.fine("Das MIF-Objekt kann nicht auf den Zieltyp '" +
			// davzieltyp + "' konvertiert werden");
			return null;
		}

		final String pidname = getPIDSpalte(pidspaltenname, mo);
		final String pid = bildePID(pidname, davtyp.substring(davtyp.indexOf(".") + 1)); //$NON-NLS-1$

		String name;
		// benutze den Text als name, wenn es ein TEXT-Objekt ist
		// Text-Objekte haben eine Punktgeometrie
		if (mo.istTextObjekt()) {
			name = mo.getText();
		} else {
			name = bildeNamen(pidname);
		}

		// if (davzieltyp != null && !davtyp.equals(davzieltyp)) {
		// davtyp = KonfigurationsDaten.DAVTYP_UNBESTIMMT;
		// }

		pids.add(pid);

		if (testeTyp(davtyp, "typ.punktXY")) {
			return new MifGeoPunkt(pid, name, davtyp, mo.getGeometrie(), mo.getFeature().getFeatureType());
		} else if (testeTyp(davtyp, "typ.linieXY")) {
			return new MifGeoLinie(pid, name, davtyp, mo.getGeometrie(), mo.getFeature().getFeatureType());
		} else if (testeTyp(davtyp, "typ.flächeXY")) {
			return new MifGeoFlaeche(pid, name, davtyp, mo.getGeometrie(), mo.getFeature().getFeatureType());
		} else if (testeTyp(davtyp, "typ.komplexXY")) {
			return new MifGeoKomplex(pid, name, davtyp, mo.getGeometrie(), mo.getFeature().getFeatureType());
		} else {
			pids.remove(pid);
			throw new ImportException("Der angeforderte Typ '" + //$NON-NLS-1$
					davtyp + "' ist nicht von einem der Typen des TeilmodellGeoReferenzierungGlobal abgeleitet"); //$NON-NLS-1$

		}

	}

	/**
	 * Erzeugt ein Konfigurationsobjekt.
	 *
	 * @param pid
	 *            PID
	 * @param name
	 *            Name
	 * @param geometrie
	 *            Geometrie
	 * @param featuretyp
	 *            FeatureTyp
	 *
	 * @return {@link KonfigurationsObjekt}
	 * @throws ImportException
	 *             bei Ausnahmen
	 */
	public KonfigurationsObjekt erzeugeMifGeoObjekt(final String pid, final String name, final Geometry geometrie,
			final FeatureType featuretyp) throws ImportException {
		final String davtyp;
		try {
			davtyp = bestimmeDavTyp(MifTyp.bestimmeTyp(geometrie), null);
		} catch (final ImportException ie) {
			throw ie;
		}

		final String korrpid = aktualisierePID(pid);
		pids.add(korrpid);

		if (testeTyp(davtyp, "typ.punktXY")) {
			return new MifGeoPunkt(korrpid, name, davtyp, geometrie, featuretyp);
		} else if (testeTyp(davtyp, "typ.linieXY")) {
			return new MifGeoLinie(korrpid, name, davtyp, geometrie, featuretyp);
		} else if (testeTyp(davtyp, "typ.flächeXY")) {
			return new MifGeoFlaeche(korrpid, name, davtyp, geometrie, featuretyp);
		} else if (testeTyp(davtyp, "typ.komplexXY")) { //$NON-NLS-1$
			if (geometrie.getNumGeometries() == 1) {
				return new MifGeoFlaeche(korrpid, name, "typ.flächeXY", geometrie, featuretyp); //$NON-NLS-1$
			}

			return new MifGeoKomplex(korrpid, name, davtyp, geometrie, featuretyp);
		} else {
			pids.remove(korrpid);
			throw new ImportException("Der angeforderte Typ '" + //$NON-NLS-1$
					davtyp + "' ist nicht von einem der Typen des TeilmodellGeoReferenzierungGlobal abgeleitet"); //$NON-NLS-1$
		}
	}

	/**
	 * Bestimmt den zu verwendenden Wert f&uuml;r die Bildung von PID und Name.
	 * Der Wert der Spalte, deren Name mit dem vorgegebenen Spaltennamen
	 * &uuml;bereinstimmt, wird verwendet. Existiert die vorgegebene Spalte
	 * nicht, oder wurden keine Vorgaben gemacht, so wird die erste Spalte
	 * verwendet.
	 *
	 * @param pidspaltenname
	 *            zu benutzende Spalte
	 * @param mifobjekt
	 *            {@link MifObjekt}
	 * @return Wert f&uuml;r PID und Name
	 */
	private String getPIDSpalte(final String pidspaltenname, final MifObjekt mifobjekt) {
		String pid = null;

		if ((pidspaltenname != null) && (pidspaltenname.length() > 0)) {
			final String sp = (String) mifobjekt.getFeature().getAttribute(pidspaltenname);
			if (sp != null) {
				pid = sp;
			}
		}

		if (pid == null) {
			// Default-Pid ist das erste Attribut des Features (die erste Spalte
			// (Attribut 0 ist
			// Geometrie))
			final Object attribute = mifobjekt.getFeature().getAttribute(1);
			pid = attribute.toString();
		}

		return pid;
	}

	/**
	 * Testet, ob ein Typ gleich einem anderen oder ein abgeleiteter Typ ist.
	 *
	 * @param davtyp
	 *            Datenverteiler-Typ
	 * @param totest
	 *            zu testender Typ
	 * @return true, wenn der zu testenden Typ gleich den Datenverteiler-Typ
	 *         oder von diesem abgeleitet ist.
	 */
	private boolean testeTyp(final String davtyp, final String totest) {
		final SystemObjectType dt = dav.getDataModel().getType(davtyp);
		final SystemObjectType stot = dav.getDataModel().getType(totest);

		return dt.equals(stot) || dt.inheritsFrom(stot);

	}

	/**
	 * Aktualisiert eine ge&auml;nderte PID.
	 *
	 * @param oldpid
	 *            alte PID
	 * @param newpid
	 *            neue PID
	 */
	public void updatePID(final String oldpid, final String newpid) {
		pids.remove(oldpid);
		pids.add(newpid);

	}
}
