/*
 * Segment 2 (KEx),  SWE 2.DE Statischer Import MIF
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.kex.mif.vew;

import java.io.File;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import de.bsvrz.dav.daf.main.ClientDavConnection;
import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.kex.mif.konfigdaten.ImportException;
import de.bsvrz.kex.mif.konfigdaten.KonfigurationsDaten;
import de.bsvrz.kex.mif.mif.MifTyp;
import de.bsvrz.sys.funclib.application.StandardApplication;
import de.bsvrz.sys.funclib.application.StandardApplicationRunner;
import de.bsvrz.sys.funclib.commandLineArgs.ArgumentList;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Hauptklasse f&uuml;r die SWE 2.MIF.
 *
 * @author BitCtrl Systems GmbH, Gieseler
 */
public class KexMIF implements StandardApplication {
	/**
	 * MAIN-Methode.
	 *
	 * @param args
	 *            Argumente
	 */
	public static void main(final String[] args) {
		Thread.setDefaultUncaughtExceptionHandler(new Thread.UncaughtExceptionHandler() {

			@Override
			public void uncaughtException(final Thread t, final Throwable e) {
				e.printStackTrace();
				Logger.getLogger(getClass().getName()).log(Level.INFO, "*** Application wird beendet ***", e); //$NON-NLS-1$
			}

		});

		String gui = null;
		final ArgumentList argumentList = new ArgumentList(args);
		if (argumentList.hasArgument(MIFProperties.PROP_KEY_WITH_GUI)) {
			gui = argumentList.fetchArgument(MIFProperties.PROP_KEY_WITH_GUI).asString();
		}
		if ((gui == null) || (gui.length() == 0) || (!gui.equalsIgnoreCase("ja"))) {
			StandardApplicationRunner.run(new KexMIF(), args);
		} else {
			StandardApplicationRunner.run(new KexMIFgui(), args);
		}
	}

	/**
	 * DebugLogger für Debug-Ausgaben.
	 */
	Debug debug;

	/**
	 * Datenverteiler-Verbindung.
	 */
	ClientDavConnection dav;

	/**
	 * Die erzeugten Konfigurationsdaten.
	 */
	protected KonfigurationsDaten konfigurationsdaten = null;

	/**
	 * Standard-Konstruktor.
	 */
	protected KexMIF() {
		super();
	}

	/**
	 * Erzeugt die Liste der zu importierenden MIF-Typen aus der Liste des
	 * Aufrufargumentes.
	 *
	 * @return Liste von zu importierenden MIF-Typen
	 */
	private List<MifTyp> bildeMifTypen() {
		final String typenamen = MIFProperties.getInstance().getProperty(MIFProperties.PROP_KEY_QUELLTYPEN);
		if ((typenamen != null) && (typenamen.length() > 0)) {
			final List<MifTyp> miftypen = new ArrayList<MifTyp>();
			for (final String name : typenamen.split(",")) {
				try {
					miftypen.add(MifTyp.bestimmeTyp(name));
				} catch (final ImportException e) {
					debug.error(e.getMessage());
				}
			}
			return miftypen;
		}
		return null;
	}

	/**
	 * &Uuml;berpr&uuml;ft, ob alle notwendigen Parameter f&uuml;r den Import
	 * gesetzt sind.
	 *
	 * @throws ImportException
	 *             bei fehlenden notwendigen Paramertern
	 */
	public final void checkParameter() throws ImportException {
		// teste notwendige Parameter
		final String[] toTest = { MIFProperties.PROP_KEY_QUELLDATEI, MIFProperties.PROP_KEY_KBPFAD,
				MIFProperties.PROP_KEY_KV, MIFProperties.PROP_KEY_KBPID, MIFProperties.PROP_KEY_KBINFO };

		for (final String element : toTest) {
			final String testprop = MIFProperties.getInstance().getProperty(element);
			if ((testprop == null) || (testprop.length() == 0)) {
				throw new ImportException("Der Parameter '" + element + "' fehlt!"); //$NON-NLS-1$//$NON-NLS-2$
			}
		}

		// teste optionale Parameter
		final String[] toWarn = { MIFProperties.PROP_KEY_KBNAME };
		for (final String element : toWarn) {
			final String testprop = MIFProperties.getInstance().getProperty(element);
			if ((testprop == null) || (testprop.length() == 0)) {
				debug.warning("Der Parameter '" + element + "' ist nicht gesetzt!"); //$NON-NLS-1$//$NON-NLS-2$
			}
		}

	}

	@Override
	public final void initialize(final ClientDavInterface connection) throws Exception {
		debug = Debug.getLogger();
		dav = (ClientDavConnection) connection;

		try {
			startImport();
		} catch (final Exception e) {
			// debug.error(e.getMessage());
			throw e;
		}

		// connection.disconnect(false, ""); //$NON-NLS-1$
	}

	/**
	 * Wertet die an die Applikation &uuml;bergebenen Kommandozeilenparameter
	 * aus.<br>
	 * Zus&auml;tzlich zu den Standard-Datenverteiler-Argumenten sind folgende
	 * Argumente m&ouml;glich:
	 * <ul>
	 * <li>-guiStatischerImportMif=&lt;boolean&gt;, Schalter f&uuml;r grafische
	 * Oberfl&auml;che [ja|nein] (Standardeinstellung: nein), legt fest, ob die
	 * SWE mit oder ohne grafischem Interface gestartet wird.</li>
	 * <li>-mifDatenQuelle=&lt;Datei&gt;, Dateiname der zu importierenden
	 * MIF-Datenquelle. Pfad/Name der einzulesenden .mif- und .mid Dateien</li>
	 * <li>-pidSpaltenName=&lt;Text&gt;, Spaltenname aus der mif-Datei zur
	 * Bildung der PID, Bezeichnung der Spalten aus dem "COLUMNS"-Header der
	 * mif-Datei, der zur Bildung der PID bzw. des Namens verwendet werden soll
	 * </li>
	 * <li>-quellObjektTypen=&lt;Text,Text,...&gt;, zu importierende
	 * Quellobjekttypen</li>
	 * <li>-zielObjektTyp=&lt;Text&gt;, zu erzeugender Zielobjekttyp</li>
	 * <li>-dateiNameKonfigurationsBereich=&lt;Datei&gt;, Dateiname f&uuml;r den
	 * zu erzeugenden Konfigurationsbereich.</li>
	 * <li>-KonfigurationsVerantwortlicher=&lt;Text&gt;,
	 * Konfigurationsverantwortlicher des zu erzeugenden Konfigurationsbereichs
	 * </li>
	 * <li>-KonfigurationsBereichsPid=&lt;Text&gt;, PID des
	 * Konfigurationsbereichs, z. B. 'kb.objekteGew&auml;sserBw'</li>
	 * <li>-KonfigurationsBereichsName=&lt;Text&gt;, Name des
	 * Konfigurationsbereichs, z. B. 'Teilmodell Gew&auml;sser, Objekte in BW'
	 * </li>
	 * <li>-KonfigurationsBereichsInfo=&lt;Text&gt;, Infotext des
	 * Konfigurationsbereichs, z. B. 'Teilmodell Gew&auml;sser, Objekte für
	 * Baden-W&uuml;rttemberg, erzeugt aus der MIF-Datei gew&auml;sser.mif'.
	 * </li>
	 * <li>-abweichungApproximation=&lt;Grad&gt;, Abweichung zwischen den
	 * St&uuml;tzpunkten bei Approximation Standardwert ist 0,000010 Grad (ca. 1
	 * Meter).</li>
	 * <li>-importAlleTypen=&lt;boolean&gt;, Schalter über den festgelegt wird,
	 * ob nicht auf den angegebenen Zieltyp konvertierbare Objekte auf die
	 * passenden Basistypen abgebildet werden (=ja, Standardeinstellung) oder ob
	 * nur die passenden Objekte importiert werden (=nein).
	 * </ul>
	 *
	 * @param argumentList
	 *            die Liste der &uuml;bergebenen Argumente
	 * @see de.bsvrz.sys.funclib.application.StandardApplication#parseArguments(de.bsvrz.sys.funclib.commandLineArgs.ArgumentList)
	 * @throws Exception
	 *             bei Ausnahmen
	 */
	@Override
	public final void parseArguments(final ArgumentList argumentList) throws Exception {
		final String[] myargs = { MIFProperties.PROP_KEY_WITH_GUI, MIFProperties.PROP_KEY_KBPFAD,
				MIFProperties.PROP_KEY_KV, MIFProperties.PROP_KEY_KBPID, MIFProperties.PROP_KEY_KBNAME,
				MIFProperties.PROP_KEY_KBINFO, MIFProperties.PROP_KEY_QUELLDATEI, MIFProperties.PROP_KEY_PIDSPALTE,
				MIFProperties.PROP_KEY_QUELLTYPEN, MIFProperties.PROP_KEY_ZIELTYP, MIFProperties.PROP_KEY_APPROXIMATION,
				MIFProperties.PROP_KEY_IMPORTALL };

		for (final String myarg : myargs) {
			if (argumentList.hasArgument(myarg)) {
				MIFProperties.getInstance().setProperty(myarg, argumentList.fetchArgument(myarg).asString());
			}
		}

	}

	/**
	 * F&uuml;hrt die Importoperation aus.
	 *
	 * @throws Exception
	 *             bei Ausnahmen.
	 */
	public void startImport() throws Exception {
		try {
			debug = Debug.getLogger();

			checkParameter();

			debug.info("Applikation wird gestartet"); //$NON-NLS-1$

			konfigurationsdaten = new KonfigurationsDaten();
			// bilde den Dateinamen den Konfigurationsbereiches aus
			// Pfad+Pid+".xml"
			final File kbfile = new File(MIFProperties.getInstance().getProperty(MIFProperties.PROP_KEY_KBPFAD),
					MIFProperties.getInstance().getProperty(MIFProperties.PROP_KEY_KBPID) + ".xml"); //$NON-NLS-1$
			konfigurationsdaten.initialisiere(dav, kbfile.toString(),
					MIFProperties.getInstance().getProperty(MIFProperties.PROP_KEY_KV),
					MIFProperties.getInstance().getProperty(MIFProperties.PROP_KEY_KBPID),
					MIFProperties.getInstance().getProperty(MIFProperties.PROP_KEY_KBNAME),
					MIFProperties.getInstance().getProperty(MIFProperties.PROP_KEY_KBINFO));
			konfigurationsdaten
			.importiereMIF(MIFProperties.getInstance().getProperty(MIFProperties.PROP_KEY_QUELLDATEI));
			konfigurationsdaten.konvertiere(MIFProperties.getInstance().getProperty(MIFProperties.PROP_KEY_PIDSPALTE),
					bildeMifTypen(), MIFProperties.getInstance().getProperty(MIFProperties.PROP_KEY_ZIELTYP));
			konfigurationsdaten.exportiereKB();
		} catch (final Exception e) {
			throw e;
		}

		debug.info("Applikation beendet"); //$NON-NLS-1$
		// _davdisconnect(false, ""); //$NON-NLS-1$
		dav.disconnect(false, "");
	}

}
